import SkribblClient from "../logic/SkribblClient.js";
import { CustomElement } from "../util/Util.js";
/**
 * @typedef {import("../logic/SkribblServer.js").DrawOrder} DrawOrder
 */



/**
 * Custom element `<skribbl-canvas>` for the canvas that can be drawn on by whoever is currently drawing.
 */
export default class SkribblCanvas extends CustomElement {
    /**
     * @param {SkribblClient} client 
     */
    constructor(client) {
        super();
        this._drawable = false;
        this.attachShadow({ mode: "open" });
        this.shadowRoot.innerHTML = `
			<style>
				:host {
					display: block;
					width: 100%;
					height: 100%;
				}
				canvas {
					width: 100%;
					height: 100%;
				}
			</style>
			<canvas width="960" height="720"></canvas>
		`;
        this._canvas = this.shadowRoot.querySelector("canvas");
        this._radius = 10;
        this._color = "red";
        this._client = client;

        /**
         * looks what kind of order is given, then either draws, erases(=drawing with background color), changes the backgroundcolor or clears the canvas
         * Todo: test functionality
         */
        client.onDraw(order => {
            // runs whenever the server tells all clients to draw something on the canvas

            this._canvas = this.shadowRoot.querySelector("canvas");
            var ctx = this._canvas.getContext("2d");
            ctx.lineCap = "round";

            if (order.action == "draw") {
                let x = 960 * order.points[0].x / 1000;
                let y = 720 * order.points[0].y / 1000;
                ctx.beginPath();
                ctx.moveTo(x, y);
                let xx = 960 * order.points[1].x / 1000;
                let yy = 720 * order.points[1].y / 1000;
                ctx.lineTo(xx, yy);
                ctx.lineWidth = order.radius;
                ctx.strokeStyle = order.color;
                ctx.stroke();
            } else if (order.action == "erase") {
                let x = 960 * order.points[0].x / 1000;
                let y = 720 * order.points[0].y / 1000;
                ctx.beginPath();
                ctx.moveTo(x, y);
                let xx = 960 * order.points[1].x / 1000;
                let yy = 720 * order.points[1].y / 1000;
                ctx.lineTo(xx, yy);
                ctx.lineWidth = order.radius;
                ctx.strokeStyle = this._backgroundColor;
                ctx.stroke();
            } else if (order.action == "changeBackgroundColor") {
                this._backgroundColor = order.color;
                ctx.fillStyle = order.color;
                ctx.fillRect(0, 0, this._canvas.width, this._canvas.height);
            } else if (order.action == "clearCanvas") {
                ctx.clearRect(0, 0, this._canvas.width, this._canvas.height);
            }

            console.log("incoming draw order:", order); // just for testing, can be removed once this has been implemented
        }, { onlyWhen: this.connected });

        /**
         * is allowed to Draw and last known mouse position
         */
        this._isDrawing = false;
        this._relativeMousePositionX = 0;
        this._relativeMousePositionY = 0;


        /**
         * Sends on MouseMove event if you are allowed to draw a draw event to the server with color, radius and erase/draw
         */
        this._canvas.addEventListener("mousemove", e => {
            if (this._drawable && this._isDrawing) {
                let oldMousePositionX = this._relativeMousePositionX;
                let oldMousePositionY = this._relativeMousePositionY;
                this.updateRelativeMousePosition(e); // Size not in Percent but in perMille (per Thousand)

                /** @type {"draw"|"erase"} */
                let actionToDo = "draw";
                if (this._erasorActive) {
                    actionToDo = "erase"
                }
                /** @type {DrawOrder} */
                let coolOrder = { action: actionToDo, color: this._color, radius: this._radius, points: [{ x: oldMousePositionX, y: oldMousePositionY }, { x: this._relativeMousePositionX, y: this._relativeMousePositionY }] };
                client.draw(coolOrder);
            }
        });

        /**
         * look if you can draw on the canvas
         */
        this._canvas.addEventListener("mousedown", e => {
            this._isDrawing = true;
            this.updateRelativeMousePosition(e);
        });

        this._canvas.addEventListener("mouseup", e => {
            this._isDrawing = false;
        });

        this._canvas.addEventListener("mouseleave", e => {
            this._isDrawing = false;
        });

    }

    /**
     * Sets the backgroundColor of the canvas
     */
    /** @param {string} backgroundColor */
    set backgroundColor(backgroundColor) {
        /** @type {import("../logic/SkribblServer.js").BackgroundOrder} */
        let coolOrder = { action: "changeBackgroundColor", color: backgroundColor };
        this._client.draw(coolOrder);
    }

    /**
     * Tells the relative mouse position on the canvas in per Mille (min0, max1000)
     */
    /** @param {MouseEvent} e */
    updateRelativeMousePosition(e) {
        this._canvas = this.shadowRoot.querySelector("canvas");
        let canvasSizeWidth = this._canvas.getBoundingClientRect().width;
        let canvasSizeHeight = this._canvas.getBoundingClientRect().height;


        this._relativeMousePositionX = Math.round(1000 * e.offsetX / canvasSizeWidth); // Size not in Percent but in perMille (per Thousand)
        this._relativeMousePositionY = Math.round(1000 * e.offsetY / canvasSizeHeight); // Size not in Percent but in perMille (per Thousand)
    }

    /**
     * Whether the canvas can currently be drawn on by this player.
     */
    set drawable(drawable) {
        if (this._drawable !== drawable) {
            this._drawable = drawable;
            // if some styles need to be changed depending on whether or not the player is currently allowed to draw, like the mouse coursor for example, this is probably the right place to do so
        }
    }

    get drawable() {
        return this._drawable;
    }

    /**
     * which radius the pen should have 
     */
    /** @param {number} radius */
    set radius(radius) {
        this._radius = radius;
    }

    /**
     * Which color the pen should have 
     */
    /** @param {string} color */
    set color(color) {
        this._color = color;
    }

    /**
     * Whether the erasor is active
     */
    /** @param {boolean} erasorActive */
    set erasorActive(erasorActive) {
        this._penActive = false;
        this._erasorActive = erasorActive;
    }

    /**
     * Whether the pen is active
     */
    /** @param {boolean} penActive */
    set penActive(penActive) {
        this._penActive = false;
        this._penActive = penActive;
    }
}
customElements.define("skribbl-canvas", SkribblCanvas);